﻿using System;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {
  using System.Collections.Generic;
  using System.Linq;
  using Common;
  using Configuration;
  using Core;
  using Encodings.IntegerVectorEncoding;

  using HeuristicLab.BenchmarkSuite;
  using HeuristicLab.BenchmarkSuite.Problems;
  using HeuristicLab.Data;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  using Instances;
  using Interpreter;
  using Optimization;
  using Parameters;
  using Persistence.Default.CompositeSerializers.Storable;

  [StorableClass]
  [Creatable(CreatableAttribute.Categories.GeneticProgrammingProblems, Priority = 180)]
  [Item("Push Problem", "")]
  public class PushProblem : SingleObjectiveBasicProblem<IntegerVectorEncoding>, IProblemInstanceConsumer<Data> {
    [Storable]
    private readonly PushConfiguration config;
    private PushInterpreterPool pool;
    private IBenchmarkSuiteDataDescriptor dataDescriptor;

    public PushProblem() {
      config = new PushConfiguration();
      pool = new PushInterpreterPool(10000, 4096, 65536, config);

      InitEvents();
      InitParameters();
      Instructions = config;
    }

    [StorableConstructor]
    public PushProblem(bool deserializing)
      : base(deserializing) {
    }

    public PushProblem(PushProblem original, Cloner cloner)
      : base(original, cloner) {
      config = cloner.Clone(original.config);
      pool = new PushInterpreterPool(65536, 1024, 65536, config);

      Instructions = config;

      this.InitEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      pool = new PushInterpreterPool(config);
      Instructions = config;

      InitEvents();
    }

    private void InitEvents() {
      config.EnabledExpressionsChanged += EnabledExpressionsChanged;
    }

    private void EnabledExpressionsChanged(object sender, EnabledExpressionsChangedEventArgs e) {
      this.Encoding.Bounds[0, 1] = config.EnabledExpressions.Count - 1;
      this.Encoding.BoundsParameter.Value[0, 1] = config.EnabledExpressions.Count - 1;
    }

    #region Parameters

    private const string EvalPushLimitParameterName = "EvalPushLimit";
    private const string EvalPushLimitParameterDescription = "This is the maximum allowed number of \"executions\" in a single top-level call to the interpreter. The execution of a single Push instruction counts as one execution, as does the processing of a single literal, as does the descent into one layer of parentheses (that is, the processing of the \"(\" counts as one execution).";
    private const string MaxProgramLengthParameterName = "MaxProgramLength";
    private const string MaxProgramLengthParameterDescription = "This is the maximum size of an item on the CODE stack, expressed as a number of points. A point is an instruction, a literal, or a pair of parentheses.";
    private const string TopLevelPushCodeParameterName = "TopLevelPushCode";
    private const string TopLevelPushCodeParameterDescription = "When TRUE (which is the default), code passed to the top level of the interpreter will be pushed onto the CODE stack prior to execution.";
    private const string TopLevelPopCodeParameterName = "TopLevelPopCode";
    private const string TopLevelPopCodeParameterDescription = "When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.";
    private const string InstructionsParameterName = "Instructions";
    private const string InstructionsParameterDescription = "Enables/Disables Instructions";
    private const string DataParameterName = "Data";
    private const string DataParameterDescription = "Program Synthesis";
    private const string MinRandomIntegerParameterName = "MinRandomInteger";
    private const string MinRandomIntegerParameterDescription = "The minimum INTEGER that will be produced as an ephemeral random INTEGER constant or from a call to INTEGER.RAND.";
    private const string MaxRandomIntegerParameterName = "MaxRandomInteger";
    private const string MaxRandomIntegerParameterDescription = "The maximum INTEGER that will be produced as an ephemeral random INTEGER constant or from a call to INTEGER.RAND.";
    private const string MinRandomFloatParameterName = "MinRandomFloat";
    private const string MinRandomFloatParameterDescription = "The minimum FLOAT that will be produced as an ephemeral random FLOAT constant or from a call to FLOAT.RAND.";
    private const string MaxRandomFloatParameterName = "MaxRandomFloat";
    private const string MaxRandomFloatParameterDescription = "The maximum FLOAT that will be produced as an ephemeral random FLOAT constant or from a call to FLOAT.RAND.";
    private const string NewErcNameProbabilityParameterName = "NewErcNameProbability";
    private const string NewErcNameProbabilityParameterDescription = "The probability that the selection of the ephemeral random NAME constant for inclusion in randomly generated code will produce a new name.";
    private const string ErcProbabilityParameterName = "ErcProbability";
    private const string ErcProbabilityParameterDescription = "The probability that the selection of a epheral random literal constant for inclusion in randomly generated code will produce a new literal.";
    private const string MaxPointsInRandomInstructionParameterName = "MaxPointsInRandomInstruction";
    private const string MaxPointsInRandomInstructionParameterDescription = "MaxPointsInRandomInstruction";
    private const string DataBoundsParameterName = "DataBounds";
    private const string MaxProgramDepthParameterName = "MaxProgramDepth";
    private const string MaxProgramDepthParameterDescription = "";

    private void InitParameters() {

      Parameters.Add(new FixedValueParameter<DataBounds>(DataBoundsParameterName));

      Parameters.Add(new FixedValueParameter<IntValue>(
        EvalPushLimitParameterName,
        EvalPushLimitParameterDescription,
        new IntValue(config.EvalPushLimit)));

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxProgramLengthParameterName,
        MaxProgramLengthParameterDescription,
        new IntValue(config.MaxPointsInProgram)));

      Parameters.Add(new FixedValueParameter<BoolValue>(
        TopLevelPushCodeParameterName,
        TopLevelPushCodeParameterDescription,
        new BoolValue(config.TopLevelPushCode)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<BoolValue>(
        TopLevelPopCodeParameterName,
        TopLevelPopCodeParameterDescription,
        new BoolValue(config.TopLevelPopCode)) { Hidden = true });

      Parameters.Add(new ValueParameter<IEnabledExpressionsConfiguration>(
        InstructionsParameterName,
        InstructionsParameterDescription));

      Parameters.Add(new FixedValueParameter<IntValue>(
        MinRandomIntegerParameterName,
        MinRandomIntegerParameterDescription,
        new IntValue(config.MinRandomInteger)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxRandomIntegerParameterName,
        MaxRandomIntegerParameterDescription,
        new IntValue(config.MaxRandomInteger)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<DoubleValue>(
        MinRandomFloatParameterName,
        MinRandomFloatParameterDescription,
        new DoubleValue(config.MinRandomFloat)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<DoubleValue>(
        MaxRandomFloatParameterName,
        MaxRandomFloatParameterDescription,
        new DoubleValue(config.MaxRandomFloat)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<PercentValue>(
        NewErcNameProbabilityParameterName,
        NewErcNameProbabilityParameterDescription,
        new PercentValue(config.NewErcNameProbability)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<PercentValue>(
        ErcProbabilityParameterName,
        ErcProbabilityParameterDescription,
        new PercentValue(config.ErcProbability)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxPointsInRandomInstructionParameterName,
        MaxPointsInRandomInstructionParameterDescription,
        new IntValue(config.MaxPointsInRandomExpression)) { Hidden = true });

      Parameters.Add(
        new FixedValueParameter<IntValue>(
          MaxProgramDepthParameterName,
          MaxProgramDepthParameterDescription,
          new IntValue(config.MaxDepth)));

      Parameters.Add(new ValueParameter<Data>(
        DataParameterName,
        DataParameterDescription));

      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count - 1;
      Encoding.Length = config.MaxPointsInProgram;
    }

    public IValueParameter<DataBounds> DataBoundsParameter
    {
      get { return (IValueParameter<DataBounds>)Parameters[DataBoundsParameterName]; }
    }

    public DataBounds DataBounds
    {
      get { return DataBoundsParameter.Value; }
      set { DataBoundsParameter.Value = value; }
    }


    public IValueParameter<IEnabledExpressionsConfiguration> InstructionsParameter
    {
      get { return (IValueParameter<IEnabledExpressionsConfiguration>)Parameters[InstructionsParameterName]; }
    }

    public IEnabledExpressionsConfiguration Instructions
    {
      get { return InstructionsParameter.Value; }
      set { InstructionsParameter.Value = value; }
    }

    public IValueParameter<Data> DataParameter
    {
      get { return (IValueParameter<Data>)Parameters[DataParameterName]; }
    }

    public Data Data
    {
      get { return DataParameter.Value; }
      set { DataParameter.Value = value; }
    }

    public IValueParameter<IntValue> EvalPushLimitParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[EvalPushLimitParameterName]; }
    }

    public int EvalPushLimit
    {
      get { return config.EvalPushLimit; }
      set
      {
        this.EvalPushLimitParameter.Value.Value = value;
        config.EvalPushLimit = value;
      }
    }

    public IValueParameter<IntValue> MaxDepthParameter
    {
      get { return (IValueParameter<IntValue>)this.Parameters[MaxProgramDepthParameterName]; }
    }

    public int MaxDepth
    {
      get { return config.MaxDepth; }
      set
      {
        this.MaxDepthParameter.Value.Value = value;
        config.MaxDepth = value;
      }
    }

    public IValueParameter<IntValue> MaxProgramLengthParameter
    {
      get { return (IValueParameter<IntValue>)this.Parameters[MaxProgramLengthParameterName]; }
    }

    public int MaxProgramLength
    {
      get { return config.MaxPointsInProgram; }
      set
      {
        this.MaxProgramLengthParameter.Value.Value = value;
        this.Encoding.LengthParameter.Value.Value = value;
        config.MaxPointsInProgram = value;
      }
    }

    public IValueParameter<BoolValue> TopLevelPushParameter
    {
      get { return (IValueParameter<BoolValue>)this.Parameters[TopLevelPushCodeParameterName]; }
    }

    public bool TopLevelPushCode
    {
      get { return config.TopLevelPushCode; }
      set
      {
        this.TopLevelPushParameter.Value.Value = value;
        config.TopLevelPushCode = value;
      }
    }

    public IValueParameter<BoolValue> TopLevelPopParameter
    {
      get { return (IValueParameter<BoolValue>)this.Parameters[TopLevelPopCodeParameterName]; }
    }

    public bool TopLevelPopCode
    {
      get { return config.TopLevelPopCode; }
      set
      {
        this.TopLevelPushParameter.Value.Value = value;
        config.TopLevelPopCode = value;
      }
    }

    public IValueParameter<IntValue> MinRandomIntegerParameter
    {
      get { return (IValueParameter<IntValue>)this.Parameters[MinRandomIntegerParameterName]; }
    }

    public int MinRandomInteger
    {
      get { return config.MinRandomInteger; }
      set
      {
        this.MinRandomIntegerParameter.Value.Value = value;
        config.MinRandomInteger = value;
      }
    }

    public IValueParameter<IntValue> MaxRandomIntegerParameter
    {
      get { return (IValueParameter<IntValue>)this.Parameters[MaxRandomIntegerParameterName]; }
    }

    public int MaxRandomInteger
    {
      get { return config.MaxRandomInteger; }
      set
      {
        this.MaxRandomIntegerParameter.Value.Value = value;
        config.MaxRandomInteger = value;
      }
    }

    public IValueParameter<DoubleValue> MinRandomFloatParameter
    {
      get { return (IValueParameter<DoubleValue>)this.Parameters[MinRandomFloatParameterName]; }
    }

    public double MinRandomFloat
    {
      get { return config.MinRandomFloat; }
      set
      {
        this.MinRandomFloatParameter.Value.Value = value;
        config.MinRandomFloat = value;
      }
    }

    public IValueParameter<DoubleValue> MaxRandomFloatParameter
    {
      get { return (IValueParameter<DoubleValue>)this.Parameters[MaxRandomFloatParameterName]; }
    }

    public double MaxRandomFloat
    {
      get { return config.MaxRandomFloat; }
      set
      {
        this.MaxRandomFloatParameter.Value.Value = value;
        config.MaxRandomFloat = value;
      }
    }

    public IValueParameter<PercentValue> NewErcNameProbabilityParameter
    {
      get { return (IValueParameter<PercentValue>)this.Parameters[NewErcNameProbabilityParameterName]; }
    }

    public double NewErcNameProbability
    {
      get { return config.NewErcNameProbability; }
      set
      {
        this.NewErcNameProbabilityParameter.Value.Value = value;
        config.NewErcNameProbability = value;
      }
    }

    public IValueParameter<PercentValue> ErcProbabilityParameter
    {
      get { return (IValueParameter<PercentValue>)this.Parameters[ErcProbabilityParameterName]; }
    }

    public double ErcProbability
    {
      get { return config.ErcProbability; }
      set
      {
        this.ErcProbabilityParameter.Value.Value = value;
        config.ErcProbability = value;
      }
    }

    public IValueParameter<IntValue> MaxPointsInRandomInstructionParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxPointsInRandomInstructionParameterName]; }
    }

    public int MaxPointsInRandomInstruction
    {
      get { return config.MaxPointsInRandomExpression; }
      set
      {
        this.MaxPointsInRandomInstructionParameter.Value.Value = value;
        config.MaxPointsInRandomExpression = value;
      }
    }

    #endregion

    public override bool Maximization { get { return false; } }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PushProblem(this, cloner);
    }

    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      const string bestSolutionResultName = "Best Solution";
      var bestQuality = Maximization ? qualities.Max() : qualities.Min();
      var bestIdx = Array.IndexOf(qualities, bestQuality);
      var bestIndividual = individuals[bestIdx];
      var solution = new PushSolution(bestIndividual.IntegerVector(), bestQuality, Data, random, config, DataBounds.TrainingRange.Start, DataBounds.TrainingRange.End);

      if (!results.ContainsKey(bestSolutionResultName)) {
        results.Add(new Result(bestSolutionResultName, solution));
      } else if (((PushSolution)results[bestSolutionResultName].Value).Quality < qualities[bestIdx]) {
        results[bestSolutionResultName].Value = solution;
      }
    }

    public override double Evaluate(Individual individual, IRandom random) {
      return PushEvaluator.Evaluate(individual, pool, random, Data, DataBounds.TrainingRange.Start, DataBounds.TrainingRange.End);
    }

    public void Load(Data data) {
      Data = data;
      BestKnownQuality = data.BestResult;
      MaxProgramLength = data.MaxSize;
      EvalPushLimit = data.EvalLimit;

      config.EnabledExpressions = (IList<string>)ExpressionTable.GetEnabledExpressionsByStackTypes((StackTypes)data.EnabledDataTypes);

      // update enabled stack types
      foreach (var stackType in ExpressionTable.StackTypeToNamesTable.Keys) {
        var enabledStackExpressions = config.EnabledExpressions.Intersect(ExpressionTable.StackTypeToNamesTable[stackType]);
        config.SetStack(stackType, enabledStackExpressions.Any());
      }

      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count - 1;
      Encoding.Length = config.MaxPointsInProgram;

      // TODO
      // data.MaxGenerations
      // data.ProgEvalBudget

      DataBounds.TrainingRange.Start = 0;
      DataBounds.TrainingRange.End = Data.OriginalTrainingCount - 1;
      DataBounds.TestRange.Start = Data.OriginalTrainingCount;
      DataBounds.TestRange.End = Data.OriginalTrainingCount + Data.OriginalTestCount;
    }
  }
}