﻿using System.Threading.Tasks;

namespace TestPooling {
  using System;
  using System.Diagnostics;
  using System.Threading;

  using TestPooling.Pool;

  class Program {
    private const int PartitionSize = 512;
    private const int ExampleCount = 200;
    private const int InstanceCount = 1024;
    private const int PopulationCount = 100;

    private static int TotalInstanceCount => PopulationCount * InstanceCount * ExampleCount;

    static void Main(string[] args) {
      Test1();
      Test1_2();
      Test3();
    }

    private static void Test1() {
      var sw = new Stopwatch();

      sw.Start();
      Parallel.For(0, PopulationCount, pc => {
        for (var e = 0; e < ExampleCount; e++) {
          using (var pool = new ManagedPool<Person>(PartitionSize, () => new Person())) {
            for (var i = 0; i < InstanceCount; i++) {
              var person = pool.Get();
              person.Age = pc + e + i;
              person.Name = string.Format("{0} - {1} - {2}", pc, e, i);
            }
          }
        }
      });
      sw.Stop();

      Console.WriteLine("1 - Duration: {0} - {1} for {2}/{3} instances", sw.ElapsedTicks, sw.Elapsed.TotalSeconds, ManagedPool<Person>.InstanceCount, TotalInstanceCount);
    }

    private static void Test1_2() {
      var sw = new Stopwatch();
      sw.Start();

      var provider = new ManagedPoolProvider<Person>(PartitionSize);
      provider.InitDummyPartition(() => new Person());

      Parallel.For(0, PopulationCount, pc => {
        for (var e = 0; e < ExampleCount; e++) {
          using (var pool = provider.CreatePool()) {
            for (var i = 0; i < InstanceCount; i++) {
              var person = pool.Get();
              person.Age = pc + e + i;
              person.Name = string.Format("{0} - {1} - {2}", pc, e, i);
            }
          }
        }
      });
      sw.Stop();

      Console.WriteLine("1_2 - Duration: {0} - {1} for {2}/{3} instances", sw.ElapsedTicks, sw.Elapsed.TotalSeconds, provider.InstanceCount, TotalInstanceCount);
    }

    private static void Test3() {
      var count = 0;
      var sw = new Stopwatch();

      sw.Start();
      Parallel.For(0, PopulationCount, pc => {
        for (var e = 0; e < ExampleCount; e++) {
          var list = new Person[InstanceCount];
          for (var i = 0; i < InstanceCount; i++) {
            list[i] = new Person {
              Age = pc + e + i,
              Name = string.Format("{0} - {1} - {2}", pc, e, i)
            };
          }

          Interlocked.Add(ref count, list.Length);
        }
      });
      sw.Stop();

      Console.WriteLine("3 - Duration: {0} - {1}", sw.ElapsedTicks, sw.Elapsed.TotalSeconds);
      Console.WriteLine("PartitionSize: {0}", count);
    }

    private static void Test2() {
      var pool = new ObjectPool<Person>(() => new Person(), 2048);

      Parallel.For(0, 10000, i => {
        for (var j = 0; j < 3000; j++) {
          var person = pool.Allocate();
          person.Age = i + j;
          person.Name = string.Format("{0} - {1}", i, j);
        }

        //pool.Free()
      });
    }
  }

  [Serializable]
  class Person {
    public string Name { get; set; }
    public int Age { get; set; }
  }
}
